//   
import java.util.*;

public class AbacusMachine {
    private int [] memory     = new int [65535];                                //  
    private int [] objectCode = new int [65535];                                //  
    private int pc = 0;                                                         //  

    public AbacusMachine (int obj []) {
        objectCode = obj;
        pc = 0;
        run ();
    }

    //  
    private void run () {
        final int LIMIT = java.lang.Integer.MAX_VALUE;                          //    
        int opcode = 0,                                                         //  
            fop    = 0,                                                         //  
            sop    = 0,                                                         //  
            count  = 0;

        System.out.printf ("ABACUS MACHINE RUNNING...\n");
        loop:
        try {
            for (count = 0; count < LIMIT; count++) {
                opcode  = objectCode [pc++];                                    //    
                switch (opcode) {
                    case OpcodeTable.CMD_ADD: fop = objectCode [pc++];
                                              memory [fop] = memory [fop] + 1;
                                              break;
                    case OpcodeTable.CMD_SUB: fop = objectCode [pc++];
                                              memory [fop] = memory [fop] - 1;
                                              if (memory [fop] < 0) {
                                                  System.err.printf ("%s%d", "Negative value at address ", fop);
                                                  break loop;
                                              }
                                              break;
                    case OpcodeTable.CMD_CLR: fop = objectCode [pc++];
                                              memory [fop] = 0;
                                              break;
                    case OpcodeTable.CMD_CPY: fop = objectCode [pc++];
                                              sop = objectCode [pc++];
                                              memory [sop] = memory [fop];
                                              break;
                    case OpcodeTable.CMD_BR : fop = objectCode [pc++];
                                              pc = fop;
                                              break;
                    case OpcodeTable.CMD_BRZ: fop = objectCode [pc++];
                                              sop = objectCode [pc++];
                                              if (memory [fop] == 0) pc = sop;
                                              break;
                    case OpcodeTable.CMD_JSR: fop = objectCode [pc++];          //    
                                              objectCode [fop] = pc;            //    
                                              pc = fop + 1;                     //    
                                              break;
                    case OpcodeTable.CMD_RET: fop = objectCode [pc++];
                                              pc = objectCode [fop];
                                              break;
                    case OpcodeTable.CMD_IN : fop = objectCode [pc++];
                                              memory [fop] = readInt (fop);
                                              break;
                    case OpcodeTable.CMD_OUT: fop = objectCode [pc++];
                                              System.out.printf ("%s%07d%s%d\n", "> [", fop, "]: ", memory [fop]);
                                              break;
                    case OpcodeTable.CMD_NOP: break;
                    case OpcodeTable.CMD_HLT: System.out.printf ("%s%d\n", "HALTED ON PC=", pc - 1);
                                              break loop;
                }
            }
        }
        catch (IndexOutOfBoundsException ex) {
            System.err.printf ("%s%d\n", "MEMORY EMPTY OR DESTROYED ON PC=", pc - 1);
        }
        System.out.printf ("%d%s\n", count, " CYCLES PROCESSED");               //    
    }

    //  
    private int readInt (int address) {
        Scanner in = new Scanner (System.in);
        System.out.printf ("%s%07d%s", "< [", address, "]: ");
        int num = in.nextInt ();
        return java.lang.Math.abs (num);
    }
}
