//   
import java.util.*;

public class Parser {
    //  
    private final int TT_LBL  =  0,                                             //  
                      TT_NAM  =  1,                                             //  
                      TT_COD  =  2,                                             // 
                      TT_NUM  =  3,                                             // 
                      TT_EOL  =  4,                                             //  
                      TT_ERR  = -1;                                             //  

    private ArrayList <String> src = new ArrayList <String> ();                 //  
    private SymbolTable         st = new SymbolTable        ();                 //  
    private int []             obj = new int [65535];                           //  
    private int passNumber   = 0,                                               //  
                errorCounter = 0,                                               //  
                lineNumber   = 0,                                               //     
                loc          = 0,                                               //  
                token        = 0,                                               //   ()
                data         = 0;                                               // 
    private String currLine  = "",                                              //    
                   value     = "";                                              // ,  

    //   
    public void init (ArrayList <String> src) {
        this.src = src;
        System.out.printf ("PARSING SOURCE FILE\n");
        parse ();
        if (errorCounter == 0) {
            st.printSymbolTable ();
            printObjectCode ();
            AbacusMachine abm = new AbacusMachine (obj);
        }
    }

    //   
    private void parse () {
        for (passNumber = 1; passNumber <= 2; passNumber++) {
            pass ();
        }
    }

    //    
    private void pass () {
        loc = 0;                                                                //   
        for (int i = 0; i < src.size (); i++) {
            lineNumber = i;                                                     //   
            currLine = src.get (i);                                             //  
            token = getToken ();                                                //     
            if (passNumber == 1) first (); else second ();                      //   
        }
    }

    //  
    private void first () {
        switch (token) {
            case TT_LBL: if (st.lookupSymbol (value)) {
                            message ("duplicated label " + value);
                            errorCounter++;
                         }
                         else {
                            st.putSymbol (value, 'L', loc);
                            token = getToken ();
                         }
                         break;
            case TT_NAM: if (st.lookupSymbol (value)) {
                            message ("duplicated name " + value);
                            errorCounter++;
                         }
                         else {
                            st.putSymbol (value, 'N', data);
                            token = getToken ();
                         }
                         break;
            case TT_COD:
            case TT_NUM: token = getToken (); break;
            case TT_ERR: errorCounter++; break;
        }
    }

    //  
    private void second () {
        switch (token) {
            case TT_LBL:
            case TT_NAM:
            case TT_COD:
            case TT_NUM: token = getToken (); break;
            case TT_ERR: errorCounter++; break;
        }
    }

    //      
    private int getToken () {
        // 
        if (currLine.contains ("#")) {
            currLine = currLine.substring (0, currLine.indexOf ("#")).trim ();
        }
        if (currLine.isEmpty ()) {
            return TT_EOL;
        }

        // 
        if (currLine.startsWith (":")) {
            String label = currLine;
            if (currLine.contains (" "))
                label = currLine.substring (1, currLine.indexOf (" "));
            else
                label = currLine.substring (1).trim ();
            label = label.trim ();
            if (label.isEmpty ()) {
                message ("label expected");
                return TT_ERR;
            }
            value = label;
            if (currLine.length () > label.length () + 1)
                currLine = currLine.substring (label.length () + 1).trim ();    //   - 
            else
                currLine = "";                                                  //    
            return TT_LBL;
        }

        //    ( = )
        if (currLine.contains ("=")) {
            value = currLine.substring (0, currLine.indexOf ("=")).trim ();     // 
            if (value.isEmpty ()) {
                message ("name expected");
                return TT_ERR;
            }
            currLine = currLine.substring (currLine.indexOf ("=") + 1).trim (); // 
            if (currLine.isEmpty ()) {
                message ("address expected");
                return TT_ERR;
            }
            if (isNumeric (currLine)) {                                         // <number> = 
                data = parseWord (currLine);
                currLine = "";
                return TT_NAM;
            }
            if (currLine.equals (".")) {                                        // <number> =  
                data = loc;
                currLine = "";
                return TT_NAM;
            }
            data = st.getValue (currLine);                                      // <number> =  
            currLine = "";
            return TT_NAM;
        }

        //  ( )
        int opcode = OpcodeTable.CMD_UND;
        if (currLine.contains (" ")) {
           value = currLine.substring (0, currLine.indexOf (" "));
           currLine = currLine.substring (currLine.indexOf (" ") + 1).trim ();
        }
        else {
           value = currLine.substring (0);
        }
        opcode = OpcodeTable.opcodeFromString (value);
        if (opcode < 0) {
            loc++;
            message ("unknown opcode " + value);
            errorCounter++;
            return TT_ERR;
        }
        switch (opcode) {
            case OpcodeTable.CMD_NOP:                                           //  
            case OpcodeTable.CMD_HLT:
                if (passNumber == 1) {
                    loc += 1;
                }
                else {
                    obj [loc++] = opcode;
                    currLine = "";
                }
                break;
            case OpcodeTable.CMD_ADD:                                           //   
            case OpcodeTable.CMD_SUB:
            case OpcodeTable.CMD_CLR:
            case OpcodeTable.CMD_BR :
            case OpcodeTable.CMD_JSR:
            case OpcodeTable.CMD_RET:
            case OpcodeTable.CMD_IN :
            case OpcodeTable.CMD_OUT:
                if (passNumber == 1) {
                    loc += 2;
                }
                else {
                    obj [loc++] = opcode;
                    if (currLine.isEmpty ()) {
                        message ("operand expected");
                        return TT_ERR;
                    }
                    obj [loc++] = getOperand (currLine);
                    currLine    = "";
                }
                break;
            case OpcodeTable.CMD_CPY:                                           //   
            case OpcodeTable.CMD_BRZ:
                if (passNumber == 1) {
                    loc += 3;
                }
                else {
                    obj [loc++] = opcode;
                    String [] args = currLine.split (",");                      //  
                    if (args.length != 2) {
                        message ("two operands expected");
                        return TT_ERR;
                    }
                    for (int i = 0; i < args.length; i++) {
                        args [i] = args [i].trim ();
                    }
                    obj [loc++] = getOperand (args [0]);
                    obj [loc++] = getOperand (args [1]);
                    currLine = "";
                }
                break;
        }

        return TT_EOL;
    }

    //  
    private int getOperand (String operand) {
        if (isNumeric (operand)) {                                              //  -
            token = TT_NUM;
            return parseWord (operand);
        }
        if (operand.equals (".")) {                                             //   
            token = TT_NUM;
            return loc;
        }
        if (st.lookupSymbol (operand)) {                                        //  -
            data = st.getValue (operand);
            token = TT_LBL;
            return data;
        }
        else {
            message ("missing or bad operand " + operand);
            token = TT_ERR;
            return token;
        }
    }

    //     
    private int parseWord (String lit) {
        int radix = 10;
        return Integer.parseInt (lit, radix);
    }

    //   
    private boolean isNumeric (String lit) {
        try {parseWord (lit);}
        catch (Exception ex) {return false;}
        return true;
    }

    //  
    private void message (String msg) {
        System.err.printf ("[%05d] %s\n", (lineNumber + 1), msg);
    }

    //   
    private void printObjectCode () {
        System.out.printf ("+---------------+\n");
        System.out.printf ("|OBJECT CODE    |\n");
        System.out.printf ("+-------+-------+\n");
        System.out.printf ("|Address|  Value|\n");
        System.out.printf ("+-------+-------+\n");
        for (int i = 0; i < loc; i++) {
            System.out.printf ("|%07d|%7d|\n", i, obj [i]);
        }
        System.out.printf ("+-------+-------+\n");
    }
}
